from .mydataset import ImageFolder
from .unaligned_data_loader import UnalignedDataLoader
from collections import Counter
import os
import torch
from torch.utils.data import DataLoader, WeightedRandomSampler

def get_loader(source_path, target_path, evaluation_path, transforms,
               batch_size=32, return_id=False, balanced=False):
    source_folder = ImageFolder(os.path.join(source_path),
                                transforms[source_path],
                                return_id=return_id)
    target_folder_train = ImageFolder(os.path.join(target_path),
                                      transform=transforms[target_path],
                                      return_paths=False, return_id=return_id)
    eval_folder_test = ImageFolder(os.path.join(evaluation_path),
                                   transform=transforms[evaluation_path],
                                   return_paths=True)
    if balanced:
        freq = Counter(source_folder.labels)
        class_weight = {x: 1.0 / freq[x] for x in freq}
        source_weights = [class_weight[x] for x in source_folder.labels]
        sampler = WeightedRandomSampler(source_weights,
                                        len(source_folder.labels))
        print("use balanced loader")
        source_loader = torch.utils.data.DataLoader(
            source_folder,
            batch_size=batch_size,
            sampler=sampler,
            drop_last=True,
            num_workers=4)
    else:
        source_loader = torch.utils.data.DataLoader(
            source_folder,
            batch_size=batch_size,
            shuffle=True,
            drop_last=True,
            num_workers=4)

    target_loader = torch.utils.data.DataLoader(
        target_folder_train,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=4)
    test_loader = torch.utils.data.DataLoader(
        eval_folder_test,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4)

    return source_loader, target_loader, test_loader, target_folder_train


def get_loader_class_inc(source_path, target_path, target_labeled_path, evaluation_path, transforms,
                         batch_size=32, return_id=False, balanced=False):
    source_folder = ImageFolder(os.path.join(source_path),
                                transforms[source_path],
                                return_id=return_id)
    target_folder_train = ImageFolder(os.path.join(target_path),
                                      transform=transforms[target_path],
                                      return_paths=False, return_id=return_id)
    target_folder_labeled = ImageFolder(os.path.join(target_labeled_path),
                                        transform=transforms[target_labeled_path],
                                        return_paths=False, return_id=return_id)
    eval_folder_test = ImageFolder(os.path.join(evaluation_path),
                                   transform=transforms[evaluation_path],
                                   return_paths=True)
    if balanced:
        freq = Counter(source_folder.labels)
        class_weight = {x: 1.0 / freq[x] for x in freq}
        source_weights = [class_weight[x] for x in source_folder.labels]
        sampler = WeightedRandomSampler(source_weights,
                                        len(source_folder.labels))
        print("use balanced loader")
        source_loader = torch.utils.data.DataLoader(
            source_folder,
            batch_size=batch_size,
            sampler=sampler,
            drop_last=True,
            num_workers=4)
    else:
        source_loader = torch.utils.data.DataLoader(
            source_folder,
            batch_size=batch_size,
            shuffle=True,
            drop_last=True,
            num_workers=4)

    target_loader = torch.utils.data.DataLoader(
        target_folder_train,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=4)
    target_labeled_loader = torch.utils.data.DataLoader(
        target_folder_labeled,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
        num_workers=1)
    test_loader = torch.utils.data.DataLoader(
        eval_folder_test,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4)

    return source_loader, target_loader, target_labeled_loader, \
           test_loader, target_folder_train


def get_loader_balanced(source_path, target_path, evaluation_path, transforms, batch_size=32):
    source_folder = ImageFolder(os.path.join(source_path),
                                transforms[source_path])
    target_folder_train = ImageFolder(os.path.join(target_path),
                                      transform=transforms[target_path],
                                      return_paths=False)
    eval_folder_test = ImageFolder(os.path.join(evaluation_path),
                                   transform=transforms[evaluation_path],
                                   return_paths=True)
    freq = Counter(source_folder.labels)
    class_weight = {x: 1.0 / freq[x] for x in freq}
    source_weights = [class_weight[x] for x in source_folder.labels]
    sampler = WeightedRandomSampler(source_weights, len(source_folder.labels))

    train_loader = UnalignedDataLoader()
    train_loader.initialize(source_folder, target_folder_train, batch_size, sampler=sampler)

    test_loader = torch.utils.data.DataLoader(
        eval_folder_test,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4)

    return train_loader, test_loader



import torch
import torch.nn.functional as F
import numpy as np
import logging

def test(step, dataset_test, filename, n_share, unk_class, G, C1, threshold):
    G.eval()
    C1.eval()
    correct = 0
    correct_close = 0
    size = 0
    class_list = [i for i in range(n_share)]
    class_list.append(unk_class)
    per_class_num = np.zeros((n_share + 1))
    per_class_correct = np.zeros((n_share + 1)).astype(np.float32)
    per_class_correct_cls = np.zeros((n_share + 1)).astype(np.float32)
    all_pred = []
    all_gt = []
    for batch_idx, data in enumerate(dataset_test):
        with torch.no_grad():
            img_t, label_t, path_t = data[0], data[1], data[2]
            img_t, label_t = img_t.cuda(), label_t.cuda()
            feat = G(img_t)
            out_t = C1(feat)
            out_t = F.softmax(out_t)
            entr = -torch.sum(out_t * torch.log(out_t), 1).data.cpu().numpy()
            pred = out_t.data.max(1)[1]
            k = label_t.data.size()[0]
            pred_cls = pred.cpu().numpy()
            pred = pred.cpu().numpy()
            
            pred_unk = np.where(entr > threshold)
            pred[pred_unk[0]] = unk_class
            all_gt += list(label_t.data.cpu().numpy())
            all_pred += list(pred)
            for i, t in enumerate(class_list):
                t_ind = np.where(label_t.data.cpu().numpy() == t)
                correct_ind = np.where(pred[t_ind[0]] == t)
                correct_ind_close = np.where(pred_cls[t_ind[0]] == i)
                per_class_correct[i] += float(len(correct_ind[0]))
                per_class_correct_cls[i] += float(len(correct_ind_close[0]))
                per_class_num[i] += float(len(t_ind[0]))
                correct += float(len(correct_ind[0]))
                correct_close += float(len(correct_ind_close[0]))
            size += k
    per_class_acc = per_class_correct / per_class_num
    close_p = float(per_class_correct_cls.sum() / per_class_num.sum())
    print(
        '\nTest set including unknown classes:  Accuracy: {}/{} ({:.0f}%)  '
        '({:.4f}%)\n'.format(
            correct, size,
            100. * correct / size, float(per_class_acc.mean())))
    output = [step, list(per_class_acc), 'per class mean acc %s'%float(per_class_acc.mean()),
              float(correct / size), 'closed acc %s'%float(close_p)]
    logger = logging.getLogger(__name__)
    logging.basicConfig(filename=filename, format="%(message)s")
    logger.setLevel(logging.INFO)
    print(output)
    logger.info(output)


def test_class_inc(step, dataset_test, name, num_class, G, C, known_class):
    G.eval()
    C.eval()
    ## Known Score Calculation.
    correct = 0
    size = 0
    per_class_num = np.zeros((num_class))
    per_class_correct = np.zeros((num_class)).astype(np.float32)
    class_list = [i for i in range(num_class)]
    for batch_idx, data in enumerate(dataset_test):
        with torch.no_grad():
            img_t, label_t, path_t = data[0], data[1], data[2]
            img_t, label_t = img_t.cuda(), label_t.cuda()
            feat = G(img_t)
            out_t = C(feat)
            out_t = F.softmax(out_t)
            pred = out_t.data.max(1)[1]
            correct += pred.eq(label_t.data).cpu().sum()
            pred = pred.cpu().numpy()
            k = label_t.data.size()[0]

            for i, t in enumerate(class_list):
                t_ind = np.where(label_t.data.cpu().numpy() == t)
                correct_ind = np.where(pred[t_ind[0]] == t)
                per_class_correct[i] += float(len(correct_ind[0]))
                per_class_num[i] += float(len(t_ind[0]))
            size += k
            label_t = label_t.data.cpu().numpy()
            if batch_idx == 0:
                label_all = label_t
                pred_all = out_t.data.cpu().numpy()
            else:
                pred_all = np.r_[pred_all, out_t.data.cpu().numpy()]
                label_all = np.r_[label_all, label_t]
    per_class_acc = per_class_correct / per_class_num
    print(
        '\nTest set including unknown classes:  Accuracy: {}/{} ({:.0f}%)  '
        '({:.4f}%)\n'.format(
            correct, size,
            100. * correct / size, float(per_class_acc.mean())))
    close_p = 100. * float(correct) / float(size)
    output = [step, "closed", list(per_class_acc), float(per_class_acc.mean()),
              "acc known %s"%float(per_class_acc[:known_class].mean()),
              "acc novel %s"%float(per_class_acc[known_class:].mean()), "acc %s"%float(close_p)]
    logger = logging.getLogger(__name__)
    logging.basicConfig(filename=name, format="%(message)s")
    logger.setLevel(logging.INFO)
    logger.info(output)
    print(output)
    return float(per_class_acc[:known_class].mean()), float(per_class_acc[known_class:].mean())


def feat_get(step, G, C1, dataset_source, dataset_target, save_path):
    G.eval()
    C1.eval()

    for batch_idx, data in enumerate(dataset_source):
        if batch_idx == 500:
            break
        with torch.no_grad():
            img_s = data[0]
            label_s = data[1]
            img_s, label_s = Variable(img_s.cuda()), \
                             Variable(label_s.cuda())
            feat_s = G(img_s)
            if batch_idx == 0:
                feat_all_s = feat_s.data.cpu().numpy()
                label_all_s = label_s.data.cpu().numpy()
            else:
                feat_s = feat_s.data.cpu().numpy()
                label_s = label_s.data.cpu().numpy()
                feat_all_s = np.r_[feat_all_s, feat_s]
                label_all_s = np.r_[label_all_s, label_s]
    for batch_idx, data in enumerate(dataset_target):
        if batch_idx == 500:
            break
        with torch.no_grad():
            img_t = data[0]
            label_t = data[1]
            img_t, label_t = Variable(img_t.cuda()), \
                             Variable(label_t.cuda())
            feat_t = G(img_t)
            if batch_idx == 0:
                feat_all = feat_t.data.cpu().numpy()
                label_all = label_t.data.cpu().numpy()
            else:
                feat_t = feat_t.data.cpu().numpy()
                label_t = label_t.data.cpu().numpy()
                feat_all = np.r_[feat_all, feat_t]
                label_all = np.r_[label_all, label_t]
    if not os.path.exists(save_path):
        os.makedirs(save_path, exist_ok=True)
    np.save(os.path.join(save_path, "save_%s_target_feat.npy" % step), feat_all)
    np.save(os.path.join(save_path, "save_%s_source_feat.npy" % step), feat_all_s)
    np.save(os.path.join(save_path, "save_%s_target_label.npy" % step), label_all)
    np.save(os.path.join(save_path, "save_%s_source_label.npy" % step), label_all_s)



from cleo.helpers import option
from cleo.commands.command import Command
# from deep_astro_uda.model.train import train
from deep_astro_uda.settings import DEFAULT_CONFIG_PATH
from deep_astro_uda.client.options import config_path
from deep_astro_uda.configs.config_functions import ConfigParser
from deep_astro_uda.data_utils.download_data import Downloader
import os

# TODO: Add docstrings.
class DemoCommand(Command):
    """
    Run demo pipeline for the client.
    """
    name = "demo"

    options = [
        config_path,
    ]
    
    #TODO: Write help command.
    help = ""

    def handle(self):        

        config_path = self.option('config-path')
        
        # parser = ConfigParser(file_path=output_path, filename="astroNN_10_train_config_open.yaml")
        parser = ConfigParser(config_file=config_path)

        output_path = os.path.join(os.getcwd(), "/deepastro_files")
        self.line(f"Running the DeepAstroUDA demo. All output will be saved here: {output_path}")

        # Download the Astro-nn data from the website. Save it in respective folders.
        download = Downloader(output_dir=os.path.join(output_path, "/data"))

        download.download_data(dataset_name="astro-nn")

        # Train the resnet model on the data using the default config file for astro-nn.
        train_parser = ConfigParser(config_path=config_path, filename="astroNN_10_train_config_open.yaml")

        # TODO: Implement function and add here.
        # Run inference using the test set.
            # Load saved model and attempt inference.
        self.line("Running training! Inference will proceed immediately after run. Training for DeepAstroUDA may take about an hour or two.")
        # train()









            




def inv_lr_scheduler(param_lr, optimizer, iter_num, gamma=10, power=0.75, init_lr=0.001,weight_decay=0.0005, max_iter=10000):
    #10000
    """Decay learning rate by a factor of 0.1 every lr_decay_epoch epochs."""
    #max_iter = 10000
    gamma = 10.0
    lr = init_lr * (1 + gamma * min(1.0, iter_num / max_iter)) ** (-power)
    i=0
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr * param_lr[i]
        i+=1
    return optimizer




import yaml
import os
from deep_astro_uda.settings import DATASET_NAMES, ASTRO_NN_CONFIG

class ConfigParser:
    def __init__(self, file_path, filename="config.yaml", config_file=None):

        self.__check_path_exists__(file_path)

        self.path = file_path
        self.filename = filename

        if config_file is None:
            self.config_file = os.path.join(self.path, self.filename)
        else:
            self.config_file = config_file

        self.config = self.parse_config()

    def parse_config(self):

        with open(self.config_file, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
        return config

    @property
    def config(self):
        return self.config
    
    def create_yaml_from_template(self, template=None):
        """
            Template are the yamls from the available experiments. Values provided are the datasets available.
        """
        with open(self.config_file, 'w') as yaml_file:
            yaml_file.write(yaml.dump(variables, default_flow_style=False))

    def __variables_from_template__(self, template_name):
        """
        Loads the variables from the template.
        """
        if template_name not in DATASET_NAMES:
            raise ValueError(f"Template name must be one of {DATASET_NAMES}")
        else:
            # if template_name == "astro-nn":
            #     return ASTRO_NN_CONFIG
            # elif template_name == "office_home":
            #     return OFFICE_HOME_CONFIG
            # elif template_name == "galaxy_zoo":
            #     return GALAXY_ZOO_CONFIG
            # elif template_name == "deep_adversaries":
            #     return DEEP_ADVERSARIES_CONFIG
            # else:
            #     return DEFAULT_CONFIG


    def __check_path_exists__(self):
        """
        Checks if the path exists. If the path doesn't exist, then create the path in the current directory.
        """

        if not os.path.exists(self.path):
            os.makedirs(self.path)



from cleo.application import Application

import sys

from deep_astro_uda.client.commands.demo_command import DemoCommand
from deep_astro_uda.client.commands.run_command import RunCommand
from deep_astro_uda.client.commands.infer_command import InferCommand

# TODO: Update using the correct commands.
# TODO: Add pydoc documentation.
dance_app = Application()
dance_app.add(DemoCommand())
dance_app.add(RunCommand())
dance_app.add(InferCommand())

def main() -> int:
    return dance_app.run()

if __name__ == '__main__':
    sys.exit(main())



import torch
import torch.nn.functional as F

def entropy(p):
    p = F.softmax(p)
    return -torch.mean(torch.sum(p * torch.log(p+1e-5), 1))

def get_losses_unlabeled(args, G, F1, im_data, im_data_bar, im_data_bar2, target, BCE, w_cons, device):
    """ Get losses for unlabeled samples."""
    feat = G(im_data)
    feat_bar = G(im_data_bar)
    feat_bar2 = G(im_data_bar2)

    output = F1(feat, reverse=True, eta=1.0)
    output_bar = F1(feat_bar, reverse=True, eta=1.0)
    prob, prob_bar = F.softmax(output, dim=1), F.softmax(output_bar, dim=1)
    
    # loss for adversarial adpative clustering
    aac_loss = advbce_unlabeled(args, target=target, feat=feat, prob=prob, prob_bar=prob_bar, device=device, bce=BCE)

    output = F1(feat)
    output_bar = F1(feat_bar)
    output_bar2 = F1(feat_bar2)

    prob = F.softmax(output, dim=1)
    prob_bar  = F.softmax(output_bar, dim=1)
    prob_bar2 = F.softmax(output_bar2, dim=1)

    max_probs, pseudo_labels = torch.max(prob.detach_(), dim=-1)
    mask = max_probs.ge(args.threshold).float()
    
    # loss for pseudo labeling
    pl_loss = (F.cross_entropy(output_bar2, pseudo_labels, reduction='none') * mask).mean()

    # loss for consistency
    con_loss = w_cons * F.mse_loss(prob_bar, prob_bar2)

    return aac_loss, pl_loss, con_loss

def advbce_unlabeled(args, target, feat, prob, prob_bar, device, bce):
    """ Construct adversarial adpative clustering loss."""
    target_ulb = pairwise_target(args, feat, target, device)
    prob_bottleneck_row, _ = PairEnum2D(prob)
    _, prob_bottleneck_col = PairEnum2D(prob_bar)
    adv_bce_loss = -bce(prob_bottleneck_row, prob_bottleneck_col, target_ulb)
    return adv_bce_loss

def pairwise_target(args, feat, target, device):
    """ Produce pairwise similarity label."""
    feat_detach = feat.detach()
    # For unlabeled data
    if target is None:
        rank_feat = feat_detach
        rank_idx = torch.argsort(rank_feat, dim=1, descending=True)
        rank_idx1, rank_idx2 = PairEnum2D(rank_idx)
        rank_idx1, rank_idx2 = rank_idx1[:, :args.topk], rank_idx2[:, :args.topk]
        rank_idx1, _ = torch.sort(rank_idx1, dim=1)
        rank_idx2, _ = torch.sort(rank_idx2, dim=1)
        rank_diff = rank_idx1 - rank_idx2
        rank_diff = torch.sum(torch.abs(rank_diff), dim=1)
        target_ulb = torch.ones_like(rank_diff).float().to(device)
        target_ulb[rank_diff > 0] = 0
    # For labeled data
    elif target is not None:
        target_row, target_col = PairEnum1D(target)
        target_ulb = torch.zeros(target.size(0) * target.size(0)).float().to(device)
        target_ulb[target_row == target_col] = 1
    else:
        raise ValueError('Please check your target.')
    return target_ulb

def PairEnum1D(x):
    """ Enumerate all pairs of feature in x with 1 dimension."""
    assert x.ndimension() == 1, 'Input dimension must be 1'
    x1 = x.repeat(x.size(0), )
    x2 = x.repeat(x.size(0)).view(-1,x.size(0)).transpose(1, 0).reshape(-1)
    return x1, x2

def PairEnum2D(x):
    """ Enumerate all pairs of feature in x with 2 dimensions."""
    assert x.ndimension() == 2, 'Input dimension must be 2'
    x1 = x.repeat(x.size(0), 1)
    x2 = x.repeat(1, x.size(0)).view(-1, x.size(1))
    return x1, x2

def sigmoid_rampup(current, rampup_length):
    """ Exponential rampup from https://arxiv.org/abs/1610.02242"""
    if rampup_length == 0:
        return 1.0
    else:
        current = np.clip(current, 0.0, rampup_length)
        phase = 1.0 - current / rampup_length
        return float(np.exp(-5.0 * phase * phase))

class BCE(nn.Module):
    eps = 1e-7
    def forward(self, prob1, prob2, simi):
        P = prob1.mul_(prob2)
        P = P.sum(1)
        P.mul_(simi).add_(simi.eq(-1).type_as(P))
        neglogP = -P.add_(BCE.eps).log_()
        return neglogP.mean()

class BCE_softlabels(nn.Module):
    """ Construct binary cross-entropy loss."""
    eps = 1e-7
    def forward(self, prob1, prob2, simi):
        P = prob1.mul_(prob2)
        P = P.sum(1)
        neglogP = - (simi * torch.log(P + BCE.eps) + (1. - simi) * torch.log(1. - P + BCE.eps))
        return neglogP.mean()




import numpy as np

try:
    import sklearn
    import sklearn.exceptions
    import sklearn.gaussian_process
except ImportError:  # pragma: no cover
    sklearn = None  # pragma: no cover

try:
    import scipy
    import scipy.optimize
except ImportError:  # pragma: no cover
    scipy = None  # pragma: no cover

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import trial as trial_module
from keras_tuner.engine import tuner as tuner_module


# Use 3 times the dimensionality of the space as the default number of
# random points.
dimensions = len(self.hyperparameters.space)
num_initial_points = self.num_initial_points or max(3 * dimensions, 3)
if len(completed_trials) < num_initial_points:
    return self._random_populate_space()

x, y = self._vectorize_trials()

# Ensure no nan, inf in x, y. GPR cannot process nan or inf.
x = np.nan_to_num(x, posinf=0, neginf=0)
y = np.nan_to_num(y, posinf=0, neginf=0)

self.gpr.fit(x, y)

def _upper_confidence_bound(x):
    x = x.reshape(1, -1)
    mu, sigma = self.gpr.predict(x, return_std=True)
    return mu - self.beta * sigma

optimal_val = float("inf")
optimal_x = None
num_restarts = 50
bounds = self._get_hp_bounds()
x_seeds = self._random_state.uniform(
    bounds[:, 0], bounds[:, 1], size=(num_restarts, bounds.shape[0])
)
for x_try in x_seeds:
    # Sign of score is flipped when maximizing.
    result = scipy.optimize.minimize(
        _upper_confidence_bound,
        x0=x_try,
        bounds=bounds,
        method="L-BFGS-B",
    )
    result_fun = (
        result.fun if np.isscalar(result.fun) else result.fun[0]
    )
    if result_fun < optimal_val:
        optimal_val = result_fun
        optimal_x = result.x

values = self._vector_to_values(optimal_x)
return {"status": trial_module.TrialStatus.RUNNING, "values": values}

def _random_populate_space(self):
        values = self._random_values()
        if values is None:
            return {"status": trial_module.TrialStatus.STOPPED, "values": None}
        return {"status": trial_module.TrialStatus.RUNNING, "values": values}

def get_state(self):
        state = super().get_state()
        state.update(
            {
                "num_initial_points": self.num_initial_points,
                "alpha": self.alpha,
                "beta": self.beta,
            }
        )
        return state

def set_state(self, state):
        super().set_state(state)
        self.num_initial_points = state["num_initial_points"]
        self.alpha = state["alpha"]
        self.beta = state["beta"]
        self.gpr = self._make_gpr()

def _vectorize_trials(self):
        x = []
        y = []
        ongoing_trials = set(self.ongoing_trials.values())
        for trial in self.trials.values():
            # Create a vector representation of each Trial's hyperparameters.
            trial_hps = trial.hyperparameters
            vector = []
            for hp in self._nonfixed_space():
                # For hyperparameters not present in the trial (either added
                # after the trial or inactive in the trial), set to default
                # value.
                if (
                    trial_hps.is_active(hp)  # inactive
                    and hp.name in trial_hps.values  # added after the trial
                ):
                    trial_value = trial_hps.values[hp.name]
                else:
                    trial_value = hp.default

                # Embed an HP value into the continuous space [0, 1].
                prob = hp.value_to_prob(trial_value)
                vector.append(prob)

    if trial in ongoing_trials:
        # "Hallucinate" the results of ongoing trials. This ensures that
        # repeat trials are not selected when running distributed.
        x_h = np.array(vector).reshape((1, -1))
        y_h_mean, y_h_std = self.gpr.predict(x_h, return_std=True)
        # Give a pessimistic estimate of the ongoing trial.
        score = y_h_mean[0] + y_h_std[0]
    elif trial.status == "COMPLETED":
        score = trial.score
        # Always frame the optimization as a minimization for
        # scipy.minimize.
        if self.objective.direction == "max":
            score = -1 * score
    elif trial.status in ["FAILED", "INVALID"]:
        # Skip the failed and invalid trials.
        continue

    x.append(vector)
    y.append(score)

x = np.array(x)
y = np.array(y)
return x, y

    def _vector_to_values(self, vector):
        hps = hp_module.HyperParameters()
        vector_index = 0
        for hp in self.hyperparameters.space:
            hps.merge([hp])
            if isinstance(hp, hp_module.Fixed):
                value = hp.value
            else:
                prob = vector[vector_index]
                vector_index += 1
                value = hp.prob_to_value(prob)

            if hps.is_active(hp):
                hps.values[hp.name] = value
        return hps.values

    def _nonfixed_space(self):
        return [
            hp
            for hp in self.hyperparameters.space
            if not isinstance(hp, hp_module.Fixed)
        ]

    def _get_hp_bounds(self):
        bounds = [[0, 1] for _ in self._nonfixed_space()]
        return np.array(bounds)


@keras_tuner_export(
    [
        "keras_tuner.BayesianOptimization",
        "keras_tuner.tuners.BayesianOptimization",
    ]
)
class Optimizer(tuner_module.Tuner):
    def __init__(
        self,
        hypermodel=None,
        objective=None,
        max_trials=10,
        num_initial_points=None,
        alpha=1e-4,
        beta=2.6,
        seed=None,
        hyperparameters=None,
        tune_new_entries=True,
        allow_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
        **kwargs
    ):
        oracle = BayesianOptimizationOracle(
            objective=objective,
            max_trials=max_trials,
            num_initial_points=num_initial_points,
            alpha=alpha,
            beta=beta,
            seed=seed,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        super().__init__(oracle=oracle, hypermodel=hypermodel, **kwargs



import os 
# Default output paths.

DEFAULT_CONFIG_PATH = f"{os.getcwd()}/deepastro_files/config_env/"

# CONFIG OPTIONS

DATASET_NAMES = ["astro-nn", "office_home", "galaxy_zoo", "deep_adversaries", "data"]

# Template configuration dictionaries.

ASTRO_NN_CONFIG = {
    "data": {},
    "model": {},
    "training": {},
    "inference": {},
    "output": {}
}

OFFICE_HOME_CONFIG = {}
GALAXY_ZOO_CONFIG = {}
DEEP_ADVERSARIES_CONFIG = {}
DEFAULT_CONFIG = {}



from cleo.helpers import option

config_path = option(
        long_name="config-path", 
        description="The path to the configuration file used to define the training, inference, and output regimen.",
        value_required=False
)

# TODO: Add descriptions.
full_options = [
        option(long_name="dataset", value_required=True),
        option(long_name="config-path", value_required=False),
        option(long_name="unknowns-supplied", value_required=True),
        option(long_name="image-path-text", value_required=False),
        option(long_name="output-directory", value_required=False),
        option(long_name="data-type", value_required=False),
        option(long_name="domain-type", value_required=False),  # implement later
    ]


from models.basenet import *
import torch


def get_model_mme(net, num_class=13, unit_size=2048, temp=0.05):
    model_g = ResBase(net, unit_size=unit_size)
    model_c = ResClassifier_MME(num_classes=num_class, input_size=unit_size, temp=temp)
    return model_g, model_c


def save_model(model_g, model_c, save_path):
    save_dic = {
        'g_state_dict': model_g.state_dict(),
        'c_state_dict': model_c.state_dict(),
    }
    torch.save(save_dic, save_path)


def load_model(model_g, model_c, load_path):
    checkpoint = torch.load(load_path)
    model_g.load_state_dict(checkpoint['g_state_dict'])
    model_c.load_state_dict(checkpoint['c_state_dict'])
    return model_g, model_c




from cleo.commands.command import Command

class InferCommand(Command):

    """
    Run inference pipeline for the client.
    """

    def handle(self):
        self.line('Running the client')
        self.call('client:infer')


from cleo.commands.command import Command
from deep_astro_uda.client.options import full_options
from deep_astro_uda.configs.config_functions import ConfigParser
from deep_astro_uda.data_utils.download_data import Downloader
from deep_astro_uda.settings import DATASET_NAMES, DEFAULT_CONFIG_PATH
import os
import shutil 
from deep_astro_uda.model.train import train
from deep_astro_uda.model.eval import test

# TODO: Add docstrings.
class RunCommand(Command):
    """
    Run training, testing, and inference pipeline.
    """
    name = "run"

    options = full_options

    help = ""

    def handle(self):
        self.line('Running the client')

        config_path = self.option('config-path')
        dataset = self.option('dataset')
        unknowns = self.option("unknowns-supplied")
        img_path = self.option("image-path-text")
        output_dir = self.option("output-directory")
        data_type = self.option("data-type")
        domain = self.option("domain-type")

        if dataset is None or dataset not in DATASET_NAMES:
            raise ValueError("You must provide a value for the dataset. Choose one of the following: astro-nn, office_home, galaxy_zoo, deep_adversaries, data.")
        elif dataset is 'data':
            # TODO:Implement check that the user-data-path (img-path) actually exists and has the right structure.
            if not os.path.exists(img_path):
                raise ValueError("Please validate that your img-path exists.")
            else:
                source_dir_path = os.path.join(img_path, "source")
                target_dir_path = os.path.join(img_path, "target")

                if not os.path.isdir(source_dir_path) and target_dir_path:
                    os.mkdir(source_dir_path)
                    os.mkdir(target_dir_path)
                    for idx, file in enumerate(os.listdir(img_path)):
                        src_path = os.path.join(file, img_path)
                        dest_path = os.path.join(file, target_dir_path)
                        if idx % 2 == 0:
                            shutil.move(src_path, dest_path)
        else:
            pass

        self.overwrite(f"The specified dataset was of type:{dataset}")

        if data_type:
            if data_type not in ['jpg', 'png', 'numpy']:
                raise ValueError("An inappropriate data type has been supplied. The only valid options are jpg, png, numpy.")
        else:
            # TODO: Implement peeking at the first image in the dataset.
            pass

        if config_path is None:
            # TODO: Implement function and add here.
            # Create the default configuration template file based on the specified dataset.
            # INSERT CREATE_CONFIG_TEMPLATE_FUNCTION at the DEFAULT_CONFIG_PATH.
            config = ConfigParser(file_path=DEFAULT_CONFIG_PATH, filename=f"{dataset}_{domain}.yaml")
            config.create_yaml_from_template(template=dataset)
        else:
            parser = ConfigParser(config_file=config_path)

        if not output_dir:
            output_path = parser.config["output_path"]
        else:
            output_path = output_dir

        self.line(f"Running the DeepAstroUDA run command. All output will be saved here: {output_path}")

        if not unknowns:
            unknowns = parser.config["nums_unknowns"]
        
        if dataset in DATASET_NAMES:
            download = Downloader(output_dir=output_path)

            download.download_data(dataset_name=dataset)


        # Input the training run.
        # dataset_test, filename, n_share, unk_class, G, C1, threshold = train()

        # # Input the inference command.
        # test(200, dataset_test, filename, n_share, unk_class, G, C1, threshold)


from __future__ import print_function
import yaml
import easydict
import os
import numpy as np
import torch.optim as optim
import torch.nn.functional as F
from torch.autograd import Variable
import torchvision.transforms as transforms
from deep_astro_uda.apex import amp, optimizers
from deep_astro_uda.model.get_loader import get_loader
from utils.utils import *
from deep_astro_uda.model.lr_schedule import inv_lr_scheduler
from deep_astro_uda.model.loss import *
from models.LinearAverage import LinearAverage
from deep_astro_uda.model.eval import test

# Training settings

from client import commands

# args = parser.parse_args()
args = commands
config_file = args.config
conf = yaml.load(open(config_file))
save_config = yaml.load(open(config_file))
conf = easydict.EasyDict(conf)
gpu_devices = ','.join([str(id) for id in args.gpu_devices])
os.environ["CUDA_VISIBLE_DEVICES"] = gpu_devices

args.cuda = torch.cuda.is_available()
source_data = args.source_path
target_data = args.target_path
evaluation_data = args.target_path

batch_size = conf.data.dataloader.batch_size
filename = source_data.split("_")[1] + "2" + target_data.split("_")[1]
filename = os.path.join("record", args.exp_name,
                        config_file.replace(".yaml", ""), filename)
if not os.path.exists(os.path.dirname(filename)):
    os.makedirs(os.path.dirname(filename))
print("record in %s " % filename)

data_transforms = {
    source_data: transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.RandomHorizontalFlip(),
        transforms.RandomCrop(224),
        transforms.ToTensor(),
        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ]),
    target_data: transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.RandomHorizontalFlip(),
        transforms.RandomCrop(224),
        transforms.ToTensor(),
        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ]),
    evaluation_data: transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.CenterCrop(224),
        transforms.ToTensor(),
        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ]),
}

use_gpu = torch.cuda.is_available()
source_loader, target_loader, \
test_loader, target_folder = get_loader(source_data, target_data,
                                        evaluation_data, data_transforms,
                                        batch_size=batch_size, return_id=True,
                                        balanced=conf.data.dataloader.class_balance)
dataset_test = test_loader
n_share = conf.data.dataset.n_share
n_source_private = conf.data.dataset.n_source_private
num_class = n_share + n_source_private

G, C1 = get_model_mme(conf.model.base_model, num_class=num_class,
                      temp=conf.model.temp)
device = torch.device("cuda")
if args.cuda:
    G.cuda()
    C1.cuda()
G.to(device)
C1.to(device)
ndata = target_folder.__len__()

## Memory
lemniscate = LinearAverage(2048, ndata, conf.model.temp, conf.train.momentum).cuda()
params = []
for key, value in dict(G.named_parameters()).items():
    if value.requires_grad and "features" in key:
        if 'bias' in key:
            params += [{'params': [value], 'lr': conf.train.multi,
                        'weight_decay': conf.train.weight_decay}]
        else:
            params += [{'params': [value], 'lr': conf.train.multi,
                        'weight_decay': conf.train.weight_decay}]
    else:
        if 'bias' in key:
            params += [{'params': [value], 'lr': 1.0,
                        'weight_decay': conf.train.weight_decay}]
        else:
            params += [{'params': [value], 'lr': 1.0,
                        'weight_decay': conf.train.weight_decay}]
criterion = torch.nn.CrossEntropyLoss().cuda()

opt_g = optim.SGD(params, momentum=conf.train.sgd_momentum,
                  weight_decay=0.0005, nesterov=True)
opt_c1 = optim.SGD(list(C1.parameters()), lr=1.0,
                   momentum=conf.train.sgd_momentum, weight_decay=0.0005,
                   nesterov=True)
[G, C1], [opt_g, opt_c1] = amp.initialize([G, C1],
                                          [opt_g, opt_c1],
                                          opt_level="O1")
G = nn.DataParallel(G)
C1 = nn.DataParallel(C1)
param_lr_g = []
for param_group in opt_g.param_groups:
    param_lr_g.append(param_group["lr"])
param_lr_f = []
for param_group in opt_c1.param_groups:
    param_lr_f.append(param_group["lr"])


def train():
    criterion = nn.CrossEntropyLoss().cuda()
    print('train start!')
    data_iter_s = iter(source_loader)
    data_iter_t = iter(target_loader)
    len_train_source = len(source_loader)
    len_train_target = len(target_loader)
    for step in range(conf.train.min_step + 1):
        G.train()
        C1.train()
        if step % len_train_target == 0:
            data_iter_t = iter(target_loader)
        if step % len_train_source == 0:
            data_iter_s = iter(source_loader)
        data_t = next(data_iter_t)
        data_s = next(data_iter_s)
        inv_lr_scheduler(param_lr_g, opt_g, step,
                         init_lr=conf.train.lr,
                         max_iter=conf.train.min_step)
        inv_lr_scheduler(param_lr_f, opt_c1, step,
                         init_lr=conf.train.lr,
                         max_iter=conf.train.min_step)
        img_s = data_s[0]
        label_s = data_s[1]
        img_t = data_t[0]
        index_t = data_t[2]
        img_s, label_s = Variable(img_s.cuda()), \
                         Variable(label_s.cuda())
        img_t = Variable(img_t.cuda())
        index_t = Variable(index_t.cuda())
        if len(img_t) < batch_size:
            break
        if len(img_s) < batch_size:
            break
        opt_g.zero_grad()
        opt_c1.zero_grad()
        ## Weight normalizztion
        C1.module.weight_norm()
        ## Source loss calculation
        feat = G(img_s)
        out_s = C1(feat)
        loss_s = criterion(out_s, label_s)

        feat_t = G(img_t)
        out_t = C1(feat_t)
        feat_t = F.normalize(feat_t)
        ### Calculate mini-batch x memory similarity
        feat_mat = lemniscate(feat_t, index_t)
        ### We do not use memory features present in mini-batch
        feat_mat[:, index_t] = -1 / conf.model.temp
        ### Calculate mini-batch x mini-batch similarity
        feat_mat2 = torch.matmul(feat_t,
                                 feat_t.t()) / conf.model.temp
        mask = torch.eye(feat_mat2.size(0),
                         feat_mat2.size(0)).bool().cuda()
        feat_mat2.masked_fill_(mask, -1 / conf.model.temp)
        loss_nc = conf.train.eta * entropy(torch.cat([out_t, feat_mat,
                                                      feat_mat2], 1))
        loss_ent = conf.train.eta * entropy_margin(out_t, conf.train.thr,
                                                   conf.train.margin)
        all = loss_nc + loss_s + loss_ent
        with amp.scale_loss(all, [opt_g, opt_c1]) as scaled_loss:
            scaled_loss.backward()
        opt_g.step()
        opt_c1.step()
        opt_g.zero_grad()
        opt_c1.zero_grad()
        lemniscate.update_weight(feat_t, index_t)
        if step % conf.train.log_interval == 0:
            print('Train [{}/{} ({:.2f}%)]\tLoss Source: {:.6f} '
                  'Loss NC: {:.6f} Loss ENS: {:.6f}\t'.format(
                step, conf.train.min_step,
                100 * float(step / conf.train.min_step),
                loss_s.item(), loss_nc.item(), loss_ent.item()))
        if step > 0 and step % conf.test.test_interval == 0:
            test(step, dataset_test, filename, n_share, num_class, G, C1,
                 conf.train.thr)
            G.train()
            C1.train()
            save_path = "record/saved_models/model_{num}.pth.tar".format(num=step)
            save_model(G, C1, save_path)


train()



from astroNN.datasets import load_galaxy10
import os
from PIL import Image
import numpy as np

class Downloader:

    def __init__(self, output_dir):
        
        self.output_dir = output_dir

    def download_data(self, dataset_name):
        """
        Downloads the specified dataset from its source. If dataset includes more than one piece, a list is returned.
        """
        if dataset_name == "astro-nn":
            self.__download_astro_nn__()
        elif dataset_name == "office_home":
            self.__download_office_home__()
        elif dataset_name == "galaxy_zoo":
            self.__download_galaxy_zoo__()
        elif dataset_name == "deep_adversaries":
            self.__download_deep_adversaries__()
        else:
            raise ValueError(f"Dataset {dataset_name} is not supported. Only the following are available: astro-nn, office_home, galaxy_zoo, deep_adversaries.")

    def __download_astro_nn__(self):
        """Comment Container. Sample - whether or not this download is taking place for the demo"""

        # Extract the images and labels.
        images,labels = load_galaxy10()

        # Create the original and simulated folders in the output directory.
        original_dir = os.path.join(self.output_dir, "original")
        simulated_dir = os.path.join(self.output_dir, "simulated")

        os.mkdir(original_dir)

        os.mkdir(simulated_dir)

        # Create the label folders in the output directory.
        self.__create_folders_from_labels__(labels, original_dir)
        self.__create_folders_from_labels__(labels, simulated_dir)

        # Save the images in their respective folders. Also, apply noise to the simulated images.
        for img, label, idx in zip(images, labels, range(len(images))):

            img_array = img.astype('uint8')
            noise = np.random.poisson(img_array * 0.5)

            # Apply the noise to the image array
            noisy_image_array = img_array + noise

            # Clip the pixel values to the valid range
            noisy_image_array = np.clip(noisy_image_array, 0, 255)

            img_out = Image.fromarray(np.uint8(noisy_image_array), 'RGB')
            if idx % 2 == 0:
                dir = original_dir
            else:
                dir = simulated_dir

            save_path = os.path.join(dir, str(label), str(idx)+'.jpg')
            img_out.save(save_path, "JPEG")
            
    def __download_office_home__(self):
        pass

    def __download_galaxy_zoo__(self):
        pass

    def __download_deep_adversaries__(self):
        pass
    
    def __create_folders_from_labels__(self, labels, directory):
        """
        Creates folders in the output directory based on the labels provided.
        """
        for label in labels:
            path = os.path.join(directory, str(label))
            if not os.path.exists(path):
                os.mkdir(path)

